//+++++++++++++++++++++++++++++++++++++++++++++++++++++//
//      Contains ENB Procedural Color Corrections      //
//           used by the Modular Shader files          //
//-----------------------CREDITS-----------------------//
// JawZ: Author and developer of this file             //
// Boris: Initial author of ENB PCC method             //
//+++++++++++++++++++++++++++++++++++++++++++++++++++++//


// This helper file is specifically only for use in the enbeffect.fx!
// The below list is only viable if the mshelpers.fxh is loaded/included into the enbeffect.fx file!


/***List of available fetches**********************************************
 * - PI // value of PI                                                    *
 * - GreyScale(_s4, IN.txcoord0.xy)                                       *
 * - AvgLuma(color.rgb).x  // or .y or .z or .w, never ever .xyzw!        *
 * - LogLuma(color.rgb)                                                   *
 * - random(uv.xy)                                                        *
 * - RGBtoXYZ(color.rgb)                                                  *
 * - XYZtoYxy(XYZ.xyz)                                                    *
 * - YxytoXYZ(XYZ.xyz, Yxy.rgb)                                           *
 * - XYZtoRGB(XYZ.xyz)                                                    *
 * - RGBToHSL(color.rgb)                                                  *
 * - HSLToRGB(hsl.rgb)                                                    *
 * - RGBtoHSV(color.rgb)                                                  *
 * - HSVtoRGB(hsv.rgb)                                                    *
 * - BlendLuma(hslbase.rgb, hslblend.rgb)                                 *
 * - SplitScreen(_s0, color, IN.txcoord0.xy, fSplitscreenPos)             *
 * - ClipMode(color.rgb)                                                  *
 **************************************************************************/


// ------------------- //
//   GUI ANNOTATIONS   //
// ------------------- //
#ifdef E_CC_PROCEDURAL
    float ECCGamma <
        string UIName="PCC: Gamma";                       string UIWidget="Spinner";  float UIMin=0.2;  float UIMax=5.0;
    > = {1.0};
    float ECCInBlack <
        string UIName="PCC: In Black";                    string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=1.0;
    > = {0.0};
    float ECCInWhite <
        string UIName="PCC: In White";                    string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=50.0;
    > = {1.0};
    float ECCOutBlack <
        string UIName="PCC: Out Black";                   string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=1.0;
    > = {0.0};
    float ECCOutWhite <
        string UIName="PCC: Out White";                   string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=1.0;
    > = {1.0};
    float ECCBrightness <
        string UIName="PCC: Brightness";                  string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=10.0;
    > = {1.0};
    float ECCContrastGrayLevel <
        string UIName="PCC: Contrast Gray Level";         string UIWidget="Spinner";  float UIMin=0.01;  float UIMax=0.99;
    > = {0.5};
    float ECCContrast <
        string UIName="PCC: Contrast";                    string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=10.0;
    > = {1.0};
    float ECCSaturation <
        string UIName="PCC: Saturation";                  string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=10.0;
    > = {1.0};
    float ECCDesaturateShadows <
        string UIName="PCC: Desaturate Shadows";          string UIWidget="Spinner";  float UIMin=0.0;  float UIMax=1.0;
    > = {0.0};

    float3 ECCColorBalanceHighlights <
        string UIName="PCC: Color Balance Highlights";    string UIWidget="Color";
    > = {0.5, 0.5, 0.5};
    float3 ECCColorBalanceShadows <
        string UIName="PCC: Color Balance Shadows";       string UIWidget="Color";
    > = {0.5, 0.5, 0.5};
    float3 ECCChannelMixerR <
        string UIName="PCC: Channel Mixer Red";           string UIWidget="Color";
    > = {1.0, 0.0, 0.0};
    float3 ECCChannelMixerG <
        string UIName="PCC: Channel Mixer Green";         string UIWidget="Color";
    > = {0.0, 1.0, 0.0};
    float3 ECCChannelMixerB <
        string UIName="PCC: Channel Mixer Blue";          string UIWidget="Color";
    > = {0.0, 0.0, 1.0};
#endif


// ------------------- //
//   HELPER CONSTANTS  //
// ------------------- //




// ------------------- //
//   HELPER FUNCTIONS  //
// ------------------- //

float3 msPCC(float3 inColor)
{
#ifdef E_CC_PROCEDURAL
  float  tempgray1;
  float3 tempgray;
  float4 tempvar;
  float3 tempcolor;

/// Levels like in photoshop, including gamma, lightness, additive brightness
    inColor = max(inColor - ECCInBlack, 0.0) / max(ECCInWhite - ECCInBlack, 0.0001);
    if (ECCGamma != 1.0) inColor = pow(inColor, ECCGamma);
    inColor = inColor * (ECCOutWhite - ECCOutBlack) + ECCOutBlack;

/// Brightness
    inColor = inColor * ECCBrightness;

/// Contrast
    inColor = (inColor - ECCContrastGrayLevel) * ECCContrast + ECCContrastGrayLevel;

/// Saturation
    tempgray = inColor;
    inColor  = lerp(tempgray, inColor, ECCSaturation);

/// Desaturate shadows
    tempvar.x  = saturate(1.0 - tempgray);
    tempvar.x *= tempvar.x;
    tempvar.x *= tempvar.x;
    inColor    = lerp(inColor, tempgray, ECCDesaturateShadows * tempvar.x);

/// Color balance
    inColor                 = saturate(inColor);
    tempgray1               = dot(inColor, 0.3333);
    float2 shadow_highlight = float2(1.0 - tempgray1, tempgray1);
    shadow_highlight       *= shadow_highlight;
    inColor.rgb            += (ECCColorBalanceHighlights * 2.0 - 1.0) * inColor * shadow_highlight.x;
    inColor.rgb            += (ECCColorBalanceShadows * 2.0 - 1.0) * (1.0 - inColor) * shadow_highlight.y;

/// Channel mixer
    tempcolor = inColor;
    inColor.r = dot(tempcolor, ECCChannelMixerR);
    inColor.g = dot(tempcolor, ECCChannelMixerG);
    inColor.b = dot(tempcolor, ECCChannelMixerB);
#endif

  return inColor;
}