//+++++++++++++++++++++++++++++++++++++++++++++++++++++//
//                   Contains Dither                   //
//           used by the Modular Shader files          //
//-----------------------CREDITS-----------------------//
// JawZ: Author and developer of this file             //
// CeeJay.dk: Author of Dither code                    //
//+++++++++++++++++++++++++++++++++++++++++++++++++++++//


// This helper file is specifically only for use in the effect.txt shader file!
// The below list is only viable if the msHelpers.fxh is loaded/included into the effect.txt file!


/***List of available fetches**********************************************
 * - PI // value of PI                                                    *
 * - DNSrSsEIFactor(WeatherAndTime.w, EXTSunrise, EXTDay, EXTSunset, EXTNight, INTSunrise, INTDay, INTSunset, INTNight)
 * - DNEIFactor(NightExt, DayExt, NightInt, DayInt)                       *
 * - DNIFactor(NightExt, DayExt, Int)                                     *
 * - DNSrSsEIFactor(_c3, WeatherAndTime.w, EXTSunrise, EXTDay, EXTSunset, EXTNight, INTSunrise, INTDay, INTSunset, INTNight) // ELE only!
 * - DNIDFactor(_c3, NightExt, DayExt, Interior, Dungeon) // ELE only!    *
 * - IDFactor(_c3, Interior, Dungeon)                     // ELE only!    *
 * - DunFactor(_c3, NightDayInt, Dungeon)                 // ELE only!    *
 * - GreyScale(_s4, IN.txcoord0.xy)                                       *
 * - AvgLuma(color.rgb).x  // or .y or .z or .w, never ever .xyzw!        *
 * - LogLuma(color.rgb)                                                   *
 * - random(uv.xy)                                                        *
 * - RGBtoXYZ(color.rgb)                                                  *
 * - XYZtoYxy(XYZ.xyz)                                                    *
 * - YxytoXYZ(XYZ.xyz, Yxy.rgb)                                           *
 * - XYZtoRGB(XYZ.xyz)                                                    *
 * - RGBToHSL(color.rgb)                                                  *
 * - HSLToRGB(hsl.rgb)                                                    *
 * - RGBtoHSV(color.rgb)                                                  *
 * - HSVtoRGB(hsv.rgb)                                                    *
 * - BlendLuma(hslbase.rgb, hslblend.rgb)                                 *
 * - SplitScreen(_s0, color, IN.txcoord0.xy, fSplitscreenPos)             *
 * - ClipMode(color.rgb)                                                  *
 **************************************************************************/


// ------------------- //
//   GUI ANNOTATIONS   //
// ------------------- //

float Empty_Row4 <
  string UIName="C--------------------------C";   string UIWidget="spinner";  float UIMin=0.0;  float UIMax=0.0;
> = {0.0};

    bool ENABLE_DITHER <
        string UIName = "Enable Dither";
    > = {false};

    bool VISUALIZE_PATTERN <
        string UIName = "Visualize Dither Pattern";
    > = {false};
    int DITHER_METHOD <
        string UIName="Dither: Choose Method";   string UIWidget="Spinner";  int UIMin=1;  int UIMax=3;
    > = {3};


// ------------------- //
//   HELPER CONSTANTS  //
// ------------------- //




// ------------------- //
//   HELPER FUNCTIONS  //
// ------------------- //

float3 msDither(float3 inColor, float2 inTexCoords)
{
  if (ENABLE_DITHER==true)
  {
    float3 color      = inColor.rgb;
    float  dither_bit = 8.0;  //Number of bits per channel. Should be 8 for most monitors.


  /// Ordered Dithering
    if (DITHER_METHOD==1)
    {
    //Calculate grid position
      float grid_position = frac(dot(inTexCoords, (ScreenSize.x * float2(1.0/16.0,10.0/36.0))) + 0.25 );

    //Calculate how big the shift should be
      float dither_shift = (0.25) * (1.0 / (pow(2, dither_bit) - 1.0));

    //Shift the individual colors differently, thus making it even harder to see the dithering pattern
      float3 dither_shift_RGB = float3(dither_shift, -dither_shift, dither_shift);  // Subpixel dithering

    //modify shift acording to grid position.
      dither_shift_RGB = lerp(2.0 * dither_shift_RGB, -2.0 * dither_shift_RGB, grid_position);  // Shift acording to grid position.

    //shift the color by dither_shift
      inColor.rgb += dither_shift_RGB;

    //Debugging features
      if (VISUALIZE_PATTERN==true && DITHER_METHOD==1)
      {
        inColor.rgb = (dither_shift_RGB * 2.0 * (pow(2,dither_bit) - 1.0) ) + 0.5;  // Visualize the RGB shift
        inColor.rgb = grid_position;  // Visualize the grid
      }
    }

  /// Random Dithering
    else if (DITHER_METHOD==2)
    {
    //Pseudo Random Number Generator
      // -- PRNG 1 - Reference --
      float seed = dot(inTexCoords, float2(12.9898,78.233));  // I could add more salt here if I wanted to
      float sine = sin(seed);                                 // cos also works well. Sincos too if you want 2D noise.
      float noise = frac(sine * 43758.5453 + inTexCoords.x);  // inTexCoords.x is just some additional salt - it can be taken out.

    //Calculate how big the shift should be
      float dither_shift = (1.0 / (pow(2,dither_bit) - 1.0));  // Using noise to determine shift. Will be 1/255 if set to 8-bit.
      float dither_shift_half = (dither_shift * 0.5);          // The noise should vary between +- 0.5
      dither_shift = dither_shift * noise - dither_shift_half; // MAD

    //shift the color by dither_shift
      inColor.rgb += float3(-dither_shift, dither_shift, -dither_shift);  // Subpixel dithering

    //Debugging features
      if (VISUALIZE_PATTERN==true && DITHER_METHOD==2)
      {
        inColor.rgb = noise;  // Visualize the noise
      }
    }

  /// New Dithering
    else if (DITHER_METHOD==3)
    {
    //Calculate grid position
      float grid_position = frac(dot(inTexCoords,(ScreenSize.x) * float2(0.75, 0.5) + (0.00025)));  // (0.6,0.8) is good too - TODO : experiment with values

    //Calculate how big the shift should be
      float dither_shift = (0.25) * (1.0 / (pow(2,dither_bit) - 1.0));              // 0.25 seems good both when using math and when eyeballing it. So does 0.75 btw.
      dither_shift = lerp(2.0 * dither_shift, -2.0 * dither_shift, grid_position);  // Shift according to grid position.

    //shift the color by dither_shift
      inColor.rgb += float3(dither_shift, -dither_shift, dither_shift);  // Subpixel dithering

    //Debugging features
      if (VISUALIZE_PATTERN==true && DITHER_METHOD==3)
      {
        inColor.rgb = grid_position;  // Visualize the grid
      }
    }
  }

  return inColor;
}